CONTENT_TYPES = {
    "js": "text/javascript",
    "css": "text/css",
    "html": "text/html",
    "png": "image/png",
    "jpg": "image/jpg",
    "jpeg": "image/jpeg",
    "gif": "image/gif",
    "svg": "image/svg+xml",
    "ico": "image/x-icon",
    "json": "application/json",
    "xml": "application/xml",
    "txt": "text/plain",
    "pdf": "application/pdf",
    "zip": "application/zip",
    "rar": "application/x-rar-compressed",
    "mp3": "audio/mpeg",
    "mp4": "video/mp4",
    "ogg": "audio/ogg",
    "ogv": "video/ogg",
    "webm": "video/webm",
    "wmv": "video/x-ms-wmv",
    "mpg": "video/mpeg",
    "mpeg": "video/mpeg",
    "avi": "video/x-msvideo",
    "webp": "image/webp",
    "ttf": "font/ttf",
    "otf": "font/otf",
    "py": "text/x-python",
}
